/* exxec.rexx - Convert ARexx script to an executable program.

   Usage: rx exxec <scriptname> [CRUNCH]

   The input file name is matched according to ARexx's own rules, in the
   following sequence:

      <scriptname>.rexx
      <scriptname>
      rexx:<scriptname>.rexx
      rexx:<scriptname>

   The output file is placed in RAM: with its ".rexx" extension, if it
   had one, removed. Warning: If your original file did NOT have a ".rexx"
   extension, and was itself in RAM:, exxec will write over it.

   The output file will be 540 or so bytes longer than the original script.
   The script itself is contained whole within the output file.

   Optionally, though, exxec will `crunch' the script by removing comments
   and multiple spaces. This is specified with the CRUNCH keyword on the
   command line (following the script file name). Depending on the coding
   style used in the script, crunching can produce substantial space savings,
   and may also make the resulting program execute very slightly faster.

   The crunching is done intelligently, with due allowance for nesting of
   both quotes and comments, and treating all quoted material as strictly
   literal. It will cause problems only for scripts that call the SOURCELINE
   built-in function or that depend on particular values of SIGL.

   The crunching process itself is quite time-consuming, increasing exxec's
   execution time by a factor of perhaps 4. If you don't care about the
   space savings, but do care about the time it takes, don't use crunching.

   =========================================================================

   Theory of operation
   -------------------

   The ARexx script (either crunched or not crunched, as described above),
   is augmented with an assembler front end that passes the script to
   ARexx as a string file to be executed as a function. Why not execute
   it as a command? Because you can't give command-line arguments to a
   string file executed as a command. However, by treating the script as
   a function to which the command line is the sole argument, the command
   line can be accessed as ARG(1) (or equivalently with ARG or PARSE ARG)
   just as though you executed it with RX.

   The script is still contained in the executable file, starting a little
   more than 500 bytes in. You can see it for yourself, if you like, with:

      type <filename> hex

   =========================================================================

   This script is freely redistributable but remains...

      Copyright (c) 1991 AHA! Software (Nick Sullivan, Chris Zamara)

****************************************************************************/


parse arg in_name    /* The source file */

crunch = 0

w = words(in_name)

if w > 1 then do
   if upper(word(in_name, w)) = 'CRUNCH' then do
      in_name = subword(in_name, 1, w - 1)
      crunch  = 1
      end
   end

call time('r')

/* Set-up:
      Read input file name; match it using same procedure as ARexx itself
      uses to match script names given to RX; create output file name; open
      the input file.
*/
if ~exists(in_name'.rexx') then
   if ~exists(in_name) then
      if ~exists('rexx:'in_name'.rexx') then
         if ~exists('rexx:'in_name) then do
            say "Unable to locate input file."
            exit
            end
         else
            in_name = 'rexx:'in_name
      else
         in_name = 'rexx:'in_name'.rexx'
   else
      nop
else
   in_name = in_name'.rexx'

parse value 'ram:'GetFileName(in_name) with out_name '.rexx'

if ~open('fpr',in_name) then do
   say "Can't open input file."
   exit
   end

/* Initialize the compound variable `out.' with the lines of hex containing
   the assembler code. The InitBinary routine returns the number of lines
   used for the assembler. We leave another line free to patch later on with
   the number of longwords used by the script itself, and start reading
   script lines and adding them from there.
*/

endbin   = InitBinary()+1        /* Make room for data-hunk-length longword */
lines    = endbin                /* Current line                            */
datasize = 0                     /* Size of script data in bytes            */
mode     = 0                     /* Code-type mode: quoted/commented/NORMAL */
incount  = 1                     /* Count of incoming lines (for report)    */


/* Start reading lines till there aren't any left, then close the file. If
   crunching is on, call CrunchLine in addition to other processing. Store
   the line, accumulate its size and loop.
*/

line = readln('fpr')
say

do while ~eof('fpr')
   say '0b'x"Processing line" incount

   if crunch then do
      line = CrunchLine(line)

      if line = '' then
         line = ''               /* Reject lines of only spaces */
      end

   if length(line) > 0 | ~crunch then do
      lines     = lines + 1
      out.lines = line'0a'x
      datasize  = datasize + length(out.lines)
      end

   line    = readln('fpr')
   incount = incount + 1
   end

call close('fpr')

/* The script must be zero-padded to a longword boundary. At least one zero
   must be present to delimit the string. The number of zeros to add is
   saved in `pad'.
*/

pad      = 4 - (datasize // 4)
datasize = (datasize + pad) % 4   /* convert padded size to longwords */

/* We have to patch hunk size information in two places: the `endbin'
   line noted after calling InitBinary(), and the 7th longword in the
   assembler part.
*/
out.endbin = d2c(datasize,4)
out.2      = overlay(out.endbin, out.2, 9)

/* We need a final line with the zero padding and `hunk end' code */
lines     = lines + 1
out.lines = copies('00'x,pad)'000003f2'x

/* At last we can write out the file */
if ~open('fpw',out_name,'w') then do
   say "Can't open output file '"out_name"'."
   exit
   end

do i=1 to lines
   call writech('fpw', out.i)
   end

call close('fpw')

say "Done! Time =" time('e') "seconds."

exit


/* GetFileName(path)

   Return file portion of path name given in arg(1).
*/
GetFileName:
   path = arg(1)
   n = lastpos("/",path)

   if n = 0 then
      n = lastpos(":",path)

   return substr(path, n + 1)


/* CrunchLine

   Compress a single line (or even less) by removing excess spaces and
   commentary. Since various exceptional conditions have to be handled,
   possibly extending over multiple lines, the behavior of this function
   is controlled by a global mode variable that corresponds to the type
   of text presently being scanned:

   Mode:    0  Normal script text
           -1  Single quoted
           -2  Double quoted
            1+ Commented (increase for each level of nesting)

   Here's what happens:

   If the current mode is non-zero, look on this line for quote mode to be
   cancelled, or for nested comment to begin or end. If none of these, return
   line unchanged if mode is < 0, else return empty line.

   If quote mode IS cancelled, split line into head and tail. Set mode to 0,
   and return head CrunchLine(tail).

   If nested comment begins or ends, change mode count accordinly and return
   CrunchLine(tail).

   If the current mode is zero, look on this line for the start of a new
   mode. If not found, return the line collapsing runs of spaces to 1.

   If the start of a new mode is found, split line into head and tail.
   Set new mode and return head CrunchLine(tail).

   NB: This would probably be more efficient if we took a more ARexx-like
   view of the source, and treated the whole script as one line line with
   semicolon terminators, and removed comments before other processing.
*/

CrunchLine: procedure expose mode

   line = arg(1)

   /* If we're already in quotes or a comment, look for the ending token
      to cancel the mode. Where two identical quotes occur in a row we
      don't want to change mode, but that's okay, we'll just pop out of
      the mode then immediately back in.
   */
   if mode ~= 0 then do
      if mode<0 then do                /* we begin within quotes     */
         qt = substr("""'",mode+3,1)   /* get quote for current mode */
         n  = 0

         do until n=0 | c              /* to EOL or find cancel  */
            n = pos(qt, line, n + 1)   /* pos'n of quote in line */
            c = n > 0
            end

         if ~c then
            return line
         else do
            mode = 0
            tail = substr(line, n + 1)

            if left(tail'.',1)=' ' then
               return left(line,n) CrunchLine(strip(tail,'l'))
            else
               return left(line, n) || CrunchLine(tail)
            end
         end
      else do
         n  = pos("*/", line)
         n2 = pos("/*", line)

         if n2>0 & n2<n then do
            mode = mode + 1
            n    = n2
            end
         else if n > 0 then
            mode = mode - 1

         if n = 0 then
            return ''
         else
            return CrunchLine(substr(line, n + 2))
         end
      end
   else do

      /* Check line for first instance of a token that would force a
         change in mode, then determine which (if any) occurs first.
      */

      tkn.1 = '"'
      tkn.2 = "'"
      tkn.3 = "/*"

      n1 = pos(tkn.1, line)
      n2 = pos(tkn.2, line)
      n3 = pos(tkn.3, line)

      if n1>0 & (n1<n2 | n2=0) & (n1<n3 | n3=0) then
         mode = -2
      else if n2>0 & (n2<n1 | n1=0) & (n2<n3 | n3=0) then
         mode = -1
      else if n3>0 & (n3<n1 | n1=0) & (n3<n2 | n2=0) then
         mode = 1
      else
         return space(line, 1)   /* Nothing special, return line as-is */

      /* Break line to right of found token */

      i = min(mode,0) + 3
      n = value('n'i) + length(tkn.i)

      parse var line head =n tail

      /* Strip the token from the end of the line */

      head = left(head, length(head) - length(tkn.i))

      /* Reserve a trailing space for head if there is at least one there
         now, and note the quote to be tacked on as well.
      */

      sp=''; if head ~== '' then if right(head,1)=' ' then sp=' '
      qt=''; if mode < 0 then qt = substr("""'", i, 1)

      return space(head, 1)sp || qt || CrunchLine(tail)
      end


/* This is the front-end - mostly the code hunk - of the eventual ARexx
   executable. I just put it down here to get it out of the way.
*/
InitBinary: procedure expose out.
   out.1 = '000003F3 00000000 00000002 00000000'x
   out.2 = '00000001 0000006F 00000010 000003E9'x
   out.3 = '0000006F 48E73F3E 24002648 284F2002'x
   out.4 = '5E400280 FFFFFFFE 9FC0244F 224A200B'x
   out.5 = '67207020 7209740A 161B6716 B6026712'x
   out.6 = 'B60067F4 B60167F0 12C3161B 6704B602'x
   out.7 = '66F64211 B5C96708 B02167F6 B21167F2'x
   out.8 = '7E142C78 00046100 00DE4A80 670000AA'x
   out.9 = '2A4043F9 0000011A 4EAEFE7A 4A806700'x
   out.10= '009243F9 0000011F 70004EAE FDD84A80'x
   out.11= '67000080 2C00CD8E 204D93C9 70004EAE'x
   out.12= 'FF706768 2648277C 00000000 0028274A'x
   out.13= '002C7002 72004EAE FF5E6746 CD8E277C'x
   out.14= '02040001 001C4EAE FF7C43F9 0000011A'x
   out.15= '4EAEFE7A 2A006708 2040224B 4EAEFE92'x
   out.16= '4EAEFF76 4A85670E 204D4EAE FE80204D'x
   out.17= '4EAEFE8C 7E00CD8E 204B7002 4EAEFF64'x
   out.18= 'CD8ECD8E 204B4EAE FF6ACD8E 22464EAE'x
   out.19= 'FE62204D 6100008C 20072E4C 4CDF7CFC'x
   out.20= '4E750000 00000000 00000000 00005245'x
   out.21= '58580072 65787873 79736C69 622E6C69'x
   out.22= '62726172 790048E7 20222C78 000470FF'x
   out.23= '4EAEFEB6 74FFB400 67402400 7022223C'x
   out.24= '00010001 4EAEFF3A 4A806728 2440157C'x
   out.25= '00040008 1542000F 93C94EAE FEDA2540'x
   out.26= '001041EA 00142548 001C43EA 00182089'x
   out.27= '200A6008 20024EAE FEB07000 4CDF4404'x
   out.28= '4E7548E7 00222448 2C780004 70FF1540'x
   out.29= '00082540 00147000 102A000F 4EAEFEB0'x
   out.30= '224A7022 4EAEFF2E 4CDF4400 4E750000'x
   out.31= '000003EC 00000003 00000000 00000060'x
   out.32= '00000070 000000B8 00000001 00000001'x
   out.33= '00000094 00000000 000003F2 000003EA'x
   return 33
